//
//
//  MineSweeper : By Gaurang Khetan
//
//  -- runs in 640X480 16-color mode

#include <STDIO.H>
#include <CONIO.H>
#include <GRAPHICS.H>
#include <STDLIB.H>
#include <TIME.H>
#include <DOS.H>

const int ALLCOLORS[3] = { YELLOW, BLUE, RED }; //according to statetype
const int GRIDCOLOR = LIGHTCYAN;
const int MINENUMCOLOR = GREEN;
const int OPENBKCOLOR = DARKGRAY;
const int CLOSEDBKCOLOR = LIGHTBLUE;
const int MINE_COLOR = RED;
const int CELL_WIDTH = 15;
const int CELL_HEIGHT = 15;

const int GAPALLOWED = 80; //time(ms) allowed betn left & right clicks
const int MAXCOLS = 30;                        // for a double click
const int MAXROWS = 24;
const int MAXMINES = 99;

enum bool { false, true };
enum statetype { opened, closed, mine_marked };
enum buttontype { leftbutton, rightbutton };

struct coordinate
	{
	int x, y;
	coordinate() {};
	coordinate(int p,int q)
		{
		x = p; y = q;
		}
	};

class crt_manager
	{
 private:
	int startx, starty;
	int no_rows, no_cols;
	int mine_number;
	time_t starttime;
	time_t timeonscrn;

	void fill_area(int,int,int,int, int);

 public:
	void set_dimen(int,int,int);
	void draw();
	void show_cell(coordinate,statetype,int=0);
	void show_mine_count(int);
	coordinate getpos(coordinate p);
	void update_mine_number(int);
	bool show_time();
	void show_mine_number();
	void show_mine_number(int);
	};

class mine_manager
	{
	int no_rows, no_cols;
	int no_mines;
	statetype** grid;
	bool** gridmine;
	int mine_correct_marked;

	int surmines(coordinate);

 public:
	void initialize(int,int,int);
	bool left_click(coordinate);
	bool right_click(coordinate);
	bool both_click(coordinate);
	};

class mouse_manager
	{
	int x,y;

 public:
	bool reset();
	void display();
	void conceal();
	void flush();
	bool was_button_pressed(buttontype);
	coordinate get_location();
	};

crt_manager screen;
mine_manager miner;
mouse_manager mice;

void main()
	{
	if (!mice.reset())
		{
		printf("Mouse not available! Press a key to terminate...");
		getch();
		exit(1);
		}

	//set graphics screen
	if (registerbgidriver(EGAVGA_driver) < 0)
		{
		printf("Error in registering");
		getch();
		exit(1);
		}


	int graph_driver, graph_mode;
	graph_driver = VGA; graph_mode = VGAHI;
	initgraph(&graph_driver, &graph_mode, "C:\\TC\\BGI");

	int t=graphresult();
	if (t != grOk)
		{
		printf("Error in init : %s ", grapherrormsg(t));
		getch();
		exit(1);
		}
	start:           //required to restart game after game is over

	cleardevice();
	gotoxy(1,1);
	printf("Hit the level: 1/2/3");
	for(;;)
	{
	char level=getch();

	if (level=='1')
		{
		//clrscr();
		screen.set_dimen(8,8,10);
		miner.initialize(8,8,10);
		break;
		}
	else if(level=='2')
		{
		//clrscr();
		screen.set_dimen(16,16,40);
		miner.initialize(16,16,40);
		break;
		}
	else if (level=='3')
		{
		screen.set_dimen(30, 16, 99);
		miner.initialize(30, 16, 99);
		break;
		}
	else
		continue;
	}
	gotoxy(1,1); printf("                              ");

	screen.draw();
	mice.display();
	mice.flush();
	coordinate point;
	do
		{
		if (screen.show_time())
			{
			gotoxy(1,1); printf(" Sorry! You were too slow...you lose!");
			sound(1000); delay(1000); nosound();
			break;
			}
		int b=0;
		if (mice.was_button_pressed(leftbutton))
			{
			delay(GAPALLOWED);
			if (mice.was_button_pressed(rightbutton))
				b = 3;
			else
				b = 1;
			}
		else if (mice.was_button_pressed(rightbutton))
			{
			delay(GAPALLOWED);
			if (mice.was_button_pressed(leftbutton))
				b = 3;
			else
				b = 2;
			}
		if (b!=0)
			{
			point = mice.get_location();
			point = screen.getpos(point);
			if (point.x == -1 && point.y == -1) b=0;
			}

		if (b==1)
			{
			bool result = miner.left_click(point);
			if (result)
				{
				gotoxy(1,1); printf("You lost");
				sound(1000); delay(1000); nosound();
				break;
				}
			}
		else if (b==2)
			{
			bool result = miner.right_click(point);
			if (result)
				{
				gotoxy(1,1); printf("You won!!");
				sound(5000); delay(1500); nosound();
				break;
				}
			}
		else if (b==3)
			{
			bool result = miner.both_click(point);
			if (result)
				{
				gotoxy(1,1); printf("You lost");
				sound(1000); delay(1000); nosound();
				break;
				}
			}

		}
	while(1);
	mice.conceal();

	while(kbhit()) getch(); // flush keyboard input
	gotoxy(1,2);
	printf("Want to play again? (y/n)");
	char ch=getch();
	gotoxy(1,2);printf("                         ");
	if (ch!='n' && ch!='N') goto start;

	closegraph();
	}

//** crt_manager member functions *//

void crt_manager::fill_area(int x1,int y1, int x2, int y2, int color)
	{
	setcolor(LIGHTRED);    //since LIGHTRED is not used elsewhere in program
	rectangle(x1, y1, x2, y2);
	setfillstyle(SOLID_FILL, color);
	floodfill(x1+1, y1+1, LIGHTRED);
	setcolor(color);
	rectangle(x1,y1,x2,y2);
	}

void crt_manager::set_dimen(int c, int r, int m)
	{
	// get them within limits
	if (c>MAXCOLS) c=MAXCOLS;
	if (r>MAXROWS) r=MAXROWS;

	//centre
	startx = 320 - (c*CELL_WIDTH)/2;
	starty = 240 - (r*CELL_HEIGHT)/2;

	//store
	no_rows = r;
	no_cols = c;
	mine_number = m;

	}

void crt_manager::draw()
	{
	int c,r,i;

	//draw grid
	setcolor(GRIDCOLOR);

	for (i=0; i<=no_cols; i++)
		line (startx+i*CELL_WIDTH,starty,
						startx+i*CELL_WIDTH,starty+no_rows*CELL_HEIGHT);
	for(i=0; i<=no_rows; i++)
		line (startx,starty+i*CELL_HEIGHT,
						startx+no_cols*CELL_WIDTH,starty+i*CELL_HEIGHT);

	//fill inside
	for ( c=0; c<=no_cols-1; c++)
		for( r=0; r<=no_rows-1; r++)
			show_cell(coordinate(c,r),closed);

	show_mine_number(mine_number);

	starttime = time(NULL);
	timeonscrn = 1000;
	show_time();
	}

void crt_manager::show_cell(coordinate p, statetype state, int no_mines)
	{
	int x = startx + CELL_WIDTH*p.x + (CELL_WIDTH/2 - 4);
	int y = starty + CELL_HEIGHT*p.y + (CELL_HEIGHT/2 - 4);

	//fill background color
	if (state==closed || state==mine_marked)
		setfillstyle(SOLID_FILL, CLOSEDBKCOLOR);
	else
		setfillstyle(SOLID_FILL, OPENBKCOLOR);
	floodfill(x,y,GRIDCOLOR);

	char text[2];
	switch(state)
		{
		case closed:					text[0] = ' '; break;
		case mine_marked:          text[0] = '#'; break;
		case opened:					text[0] = (no_mines==-1)? '*' : no_mines+48;
		}
	text[1] = 0;

	if (text[0] != 48)
		{
		settextstyle(DEFAULT_FONT,HORIZ_DIR, 0);	//also sizes letters as 8X8
		settextjustify(LEFT_TEXT, TOP_TEXT);
		setcolor( (text[0] == '*')? MINE_COLOR : ALLCOLORS[state] );
		outtextxy(x,y,text);
		}
	}

coordinate crt_manager::getpos(coordinate p)
	{
	if ( (p.y>=starty) &&  (p.y<(starty+no_rows*CELL_HEIGHT))
		&& (p.x>=startx) && (p.x<(startx+no_cols*CELL_WIDTH)) )
		return coordinate ( (p.x-startx)/CELL_WIDTH, (p.y-starty)/CELL_HEIGHT );
	else
		return coordinate(-1,-1);

	}

void crt_manager::update_mine_number(int num)
	{
	mine_number+=num;
	}

void crt_manager::show_mine_number()
	{
	show_mine_number(mine_number);
	}

void crt_manager::show_mine_number(int num)
	{
	char numtext[4];
	sprintf(numtext, "%03d", (num>0) ? num : 0 );

	fill_area(startx-1, starty-21, startx+49, starty-2, BLACK);//clean prev num

	setcolor(MINENUMCOLOR);
	settextstyle(DEFAULT_FONT,HORIZ_DIR, 2);	//also sizes letters as 16X16
	outtextxy(startx, starty-20, numtext);

	}

bool crt_manager::show_time()
	{
	time_t t;
	if ( (t = time(NULL)-starttime) == timeonscrn )
		return false;
	else
		timeonscrn = t;
	if (timeonscrn == 1000) return true;

	char text[4];
	sprintf(text, "%03d", timeonscrn);

	fill_area(startx+(no_cols*CELL_WIDTH)-49, starty-21,
			startx+(no_cols*CELL_WIDTH), starty-2, BLACK);
	setcolor(MINENUMCOLOR);
	settextstyle(DEFAULT_FONT,HORIZ_DIR, 2);	//also sizes letters as 16X16
	outtextxy(startx+(no_cols*CELL_WIDTH)-48, starty-20,text);

	return false;
	}
//** mine_manager class member functions **//

void mine_manager::initialize(int c, int r, int m)
	{
	int i,j,k;
	no_cols = (c>MAXCOLS)? MAXCOLS : c;
	no_rows = (r>MAXROWS)? MAXROWS : r;
	no_mines = (m>MAXMINES)? MAXMINES : m;
	mine_correct_marked = 0;

	//dynamically allocate array
	grid = new statetype*[no_cols];
	gridmine = new bool*[no_cols];
	for (i=0; i<=no_cols-1; i++)
		{
		grid[i] = new statetype[no_rows];
		gridmine[i] = new bool[no_rows];
		}

	//initialize array
	for(i=0; i<=no_cols-1; i++)
		for(j=0; j<=no_rows-1;j++)
			{
			grid[i][j] = closed;
			gridmine[i][j] = false;
			}

	//insert random mines into it
	randomize();
	for(k=1; k<=no_mines; k++)
		{
		do
			{
			i = random(no_cols);
			j = random(no_rows);
			}
		while(gridmine[i][j]);
		gridmine[i][j] = true;
		}

	}

int mine_manager::surmines(coordinate p)
	{
	int count=0;
	for (int i=p.x-1; i<=p.x+1; i++)
		for(int j=p.y-1; j<=p.y+1; j++)
			if (j>=0 && j<no_rows && i>=0 && i<no_cols)
				count += (gridmine[i][j])? 1 : 0;
	return count;
	}

bool mine_manager::left_click(coordinate p)
	{
	if (grid[p.x][p.y]==opened) return false;
	if (grid[p.x][p.y]==mine_marked) return false;

	grid[p.x][p.y] = opened;

	mice.conceal();

	if (gridmine[p.x][p.y])
		{
		screen.show_cell(p,grid[p.x][p.y], -1);
		mice.display();
		return true;
		}

	//count the neighbouring mines
	int count = surmines(p);

	screen.show_cell(p,grid[p.x][p.y],count);

	if (count==0)
		{
		for (int i=p.x-1; i<=p.x+1; i++)
			for(int j=p.y-1; j<=p.y+1; j++)
				if (j>=0 && j<no_rows && i>=0 && i<no_cols)
					left_click(coordinate(i,j));

		}

	sound(2000); delay(50); nosound();

	mice.display();
	return false;
	}

bool mine_manager::right_click(coordinate p)
	{
	static int mode=0;
	if ( (kbhit()) && (getch()=='r')) mode = (mode==1)? 0: 1;

	if (grid[p.x][p.y]==opened) return false;

	if (grid[p.x][p.y]==mine_marked)
		{
		grid[p.x][p.y] = closed;
		if (gridmine[p.x][p.y]) mine_correct_marked--;
		screen.update_mine_number(+1);
		if (mode==1)
			screen.show_mine_number(no_mines-mine_correct_marked);
		else if (mode==0)
			screen.show_mine_number();

		}
	else
		{
		grid[p.x][p.y]= mine_marked;
		if (gridmine[p.x][p.y]) mine_correct_marked++;
		screen.update_mine_number(-1);
		if (mode==1)
			screen.show_mine_number(no_mines-mine_correct_marked);
		else if (mode==0)
			screen.show_mine_number();

		}

	mice.conceal();
	screen.show_cell(p, grid[p.x][p.y], gridmine[p.x][p.y]);
	mice.display();

	return (mine_correct_marked==no_mines)? true : false;
	}

bool mine_manager::both_click(coordinate p)
	{
	if (grid[p.x][p.y] != opened) return false;

	int real_mines = surmines(p);
	if (real_mines==0) return false;

	int marked_mines=0;
	int i,j;

	//check for number of marked mines around it
	for(i=p.x-1; i<=p.x+1; i++)
		for(j=p.y-1; j<=p.y+1; j++)
			if (j>=0 && j<no_rows && i>=0 && i<no_cols)
				marked_mines += (grid[i][j]==mine_marked)? 1 : 0;

	if (real_mines != marked_mines) return false;

	for(i=p.x-1; i<=p.x+1; i++)
		for(j=p.y-1; j<=p.y+1; j++)
			if (j>=0 && j<no_rows && i>=0 && i<no_cols)
				if (grid[i][j] != mine_marked)
					if (left_click(coordinate(i,j)))
						return true;

	return false;
	}


//* mouse_manager member funtions *//

bool mouse_manager::reset()
	{
	_AX = 0;

	asm int 0x33;

	return (_AX==0xFFFF)? true : false;

	}

void mouse_manager::display()
	{
	_AX = 01;

	asm int 0x33;
	}

void mouse_manager::conceal()
	{
	_AX = 02;
	asm int 0x33;
	}

bool mouse_manager::was_button_pressed(buttontype button)
	{
	_AX = 5;
	_BX = button;
	asm int 0x33;
	if (_BX)
		{
		x = _CX; y = _DX;
		return true;
		}
	else
		return false;
	}

coordinate mouse_manager::get_location()
	{
	return coordinate(x,y);
	}

void mouse_manager::flush()
	{
	was_button_pressed(leftbutton);
	was_button_pressed(rightbutton);
	}